import csv
import sys
from collections import namedtuple, OrderedDict
import subprocess
import shutil
from datetime import datetime
import os
import pickle

import scripts.instanceInfo as instanceInfo


dateformat = "%Y-%m-%dT%H:%M:%S"

algorithmFields = ["isBound", "alg", "params"]
resultFields = ["mkspan", "time", "date"]

# Warning: it is important to provide the 'params' field in alphabetic
#  order, so that the system can identify the results afterwards
Algorithm = namedtuple("Algorithm", algorithmFields)
Result = namedtuple("Result", resultFields)

def makeAlgorithm(row):
    return Algorithm(**{k:row[k] for k in algorithmFields})

def makeResult(row):
    return Result(**{k:row[k] for k in resultFields})

def makeAlgorithms(name, *args, isBound=False):
    params = [{}]
    for (k,v) in args:
        params = [OrderedDict(d, **{k:x}) for x in v for d in params]

    return [ Algorithm(alg = name, isBound = isBound, params = ":".join('='.join(t) for t in p.items())) for p in params ]
    
def MakeStorage(instanceParameters, getFilePrefix, commandArgs):
    class Storage: 
        instanceFields = ["m", "k", "insType"] + instanceParameters
        questionFields = instanceFields + algorithmFields
        recordFields = instanceFields + algorithmFields + resultFields

        Instance  = namedtuple("Instance", instanceFields)
        Question  = namedtuple("Question", questionFields)
        Record = namedtuple("Record", recordFields)

        @classmethod
        def makeInstance(cls, row):
            return cls.Instance(**{k:row[k] for k in cls.instanceFields})
        @classmethod
        def makeQuestion(cls, row): 
            return cls.Question(**{k:row[k] for k in cls.questionFields})
        @classmethod
        def makeRecord(cls, row): 
            return cls.Record(**{k:row[k] for k in cls.recordFields})

        @classmethod
        def makeInstances(cls, t, **kwargs):
            result = [{"insType":t}]
            for k,v in kwargs.items():
                result = [dict(d, **{k:x}) for x in v for d in result]
                
            return list(map(cls.makeInstance, result))

        @classmethod
        def allCombinations(cls, instances, algorithms):
            return (cls.Question(**i._asdict(), **a._asdict()) for i in instances for a in algorithms)


        
        @classmethod
        def convertRow(cls, row):
            row["m"] = int(row["m"])
            row["k"] = int(row["k"])
            row["date"] = datetime.strptime(row["date"], dateformat)
            row["isBound"] = True if row["isBound"] == "True" else False

        @classmethod
        def parseRow(cls, row):
            cls.convertRow(row)
            return (cls.makeQuestion(row), makeResult(row))

        @classmethod
        def parseRowToRecord(cls, row):
            cls.convertRow(row)
            return cls.makeRecord(row)


        def __init__(self, filename):
            self.filename = filename

        @classmethod
        def getSrcFilename(cls, i):
            return getFilePrefix(i) + ".src"

        @classmethod
        def getFilename(cls, i):
            prefix = getFilePrefix(i)
            suffix = "m=%d_k=%d" % (i.m, i.k)
            return prefix + "_" + suffix + ".pm"

        @classmethod
        def updateInstanceFile(cls, i):
            src = cls.getSrcFilename(i)
            dst = cls.getFilename(i)
            dstDesc = open(dst, 'w')
            srcDesc = open(src, 'r')
            srcDesc.readline()
            print("[%s, %s]" % (i.m, i.k), file=dstDesc)
            shutil.copyfileobj(srcDesc, dstDesc)

        @classmethod
        def isInstanceFileValid(cls, i): 
            src = cls.getSrcFilename(i)
            dst = cls.getFilename(i)
            try: 
                srcTime = datetime.fromtimestamp(os.path.getmtime(src))
            except OSError as e:
                print("Instance source file %s is not readable: %s\n" % (src, e), file=sys.stderr)
                exit
            try:
                dstTime = datetime.fromtimestamp(os.path.getmtime(dst))
            except OSError:
                return False
            return dstTime >= srcTime

        @classmethod
        def ensureFilesAvailable(cls, instances):
            for i in instances:
                if not cls.isInstanceFileValid(i):
                    cls.updateInstanceFile(i)


        def readFile(self):
            allData = {}
            try: 
                with open(self.filename, 'r') as fd:
                    for row in csv.DictReader(fd, fieldnames = self.recordFields):
                        ### print(row)
                        (question, result) = self.parseRow(row)
                        if question in allData: 
                            allData[question].append(result)
                        else:
                            allData[question] = [result]
            except OSError:
                pass
            return allData

        def readRecords(self):
            with open(self.filename, 'r') as fd:
                for row in csv.DictReader(fd, fieldnames = self.recordFields):
                    record = self.parseRowToRecord(row)
                    yield record

        def _writeRecord(self, csvWriter, record):
            d = record._asdict()
            d["date"] = d["date"].strftime(dateformat)
            csvWriter.writerow(d)

        def writeRecords(self, results, overwrite=False):
            written = 0
            with open(self.filename, 'w' if overwrite else 'a') as fd:
                w = csv.DictWriter(fd, fieldnames = self.recordFields)
                for r in results:
                    self._writeRecord(w, r)
                    written += 1
            print("Written", written, "records")

        @classmethod
        def hasUptoDateAnswer(cls, q, data):
            if q in data:
                lastAnswer = max(data[q], key=lambda r:r.date)
                src = cls.getSrcFilename(q)
                try: 
                    srcTime = datetime.fromtimestamp(os.path.getmtime(src))
                except OSError as e:
                    print("Instance source file %s is not readable: %s\n" % (src, e), file=sys.stderr)
                    exit
                return lastAnswer.date >= srcTime
            else:
                return False

        @classmethod
        def hasValidAnswer(cls, q, data):
            return q in data

        def updateFile(self, questions, pred = lambda q, data: not Storage.hasValidAnswer(q, data)):
            data = self.readFile()
            ### print("Data read : \n")
            ### print(data)
            toRun = [ q for q in questions if pred(q, data)]
            ### print("\n To be run: \n")
            ## print(toRun)
            results = self.runMany(toRun)
        ###    print("\n results: \n")
        ###    print(results)

        # Keep only the last answer for each question
        def cleanupFile(self, toBeRemoved = lambda x: False):
            data = self.readFile()
            toWrite = [self.Record(*q, *max(data[q], key=lambda r:r.date)) for q in data ]
            self.writeRecords((r for r in toWrite if not toBeRemoved(r)), overwrite=True)

        def runMany(self, questions):
            separated = [ (self.makeInstance(q._asdict()), makeAlgorithm(q._asdict())) for q in questions ]

            ## Was a good idea, but pmtool has too many memory problems to open so many instances at once.
            ## Get list of instances
            # allInstances = set(i for (i,a) in separated)
            ## Group instances which require the same algorithms
            # instancesPerAlgList = {}
            # for i in allInstances:
            #     algs = tuple(a for (j, a) in separated if i == j)
            #     if algs in instancesPerAlgList:
            #         instancesPerAlgList[algs].append(i)
            #     else:
            #         instancesPerAlgList[algs] = [i]

            # results = results + sum((self.runOne(instances, algs) for (algs, instances) in instancesPerAlgList.items()), [])

            groupedByInstances = {}
            for (i, a) in separated:
                if i in groupedByInstances:
                    groupedByInstances[i].append(a)
                else:
                    groupedByInstances[i] = [a]

            nbInstances = len(groupedByInstances)
            results = sum( (self.runOne([i], algs, progress = (1+index)/nbInstances) for (index, (i, algs)) in enumerate(groupedByInstances.items())), [])

            return results


        def runOne(self, instances, algs, progress=None):
            with open(self.filename, 'a') as output:

                writer = csv.DictWriter(output, fieldnames = self.recordFields)

                self.ensureFilesAvailable(instances)
                filenames = { self.getFilename(i):i for i in instances }
                args = []
                args.extend(commandArgs)
                args.extend(["--output-raw", "--no-header"])
                args.extend(filenames.keys())
                args.extend(sum([[("-b" if a.isBound else "-a"),
                                  a.alg+(":"+a.params if a.params else "")] for a in algs], []))
                print("Running" + (" [%.1f%%] " % (progress*100) if progress else "") + ": " + " ".join(args))
                with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True) as child:
                    reader = csv.reader(child.stdout, delimiter=' ')
                    result = []
                    for row in reader:
                        ### print(row)
                        algNameAndParams=row[3].split(":", maxsplit=1)
                        rec = self.Record(*list(filenames[row[0]]), True if row[2] == 'True' else False, algNameAndParams[0],
                                          algNameAndParams[1] if len(algNameAndParams) >= 2 else "",
                                          row[4], row[5], datetime.now().strftime(dateformat))
                        writer.writerow(rec._asdict())
                        output.flush()
                        result.append(rec)
                    returncode = child.wait()
                    if returncode != 0:
                        print("Wrong return code %d for command:\n  %s\n" % (returncode, " ".join(args)), file=sys.stderr)
                        print(child.stderr, file=sys.stderr)
                        raise Exception
            return result

        def countTasks(self, instances, file=""):
            if file:
                fd = open(file, "w")
            else:
                fd = sys.stdout
            filenames = { self.getSrcFilename(i):i for i in instances } # This automatically removes duplicates
            
            counts = instanceInfo.countTasksMany(filenames.keys())

            print("instance", " ".join(instanceParameters), "count", file=fd)
            for (file, count) in counts:
                r = filenames[file]
                print(r.insType, 
                      " ".join(getattr(r, p) if getattr(r, p) else "NA" for p in instanceParameters),
                      count, file=fd)
                
        

        def printRecords(self, records, file=""):
            if file:
                fd = open(file, "w")
            else:
                fd = sys.stdout
            recordsWithParamDicts = [ r._replace(params=OrderedDict(eq.split('=') for eq in r.params.split(":")) if r.params else {}) for r in records ]
            allParams = frozenset.union(*(frozenset(r.params.keys()) for r in recordsWithParamDicts))
            print("instance",
                  " ".join(instanceParameters),
                  "m k isBound algorithm",
                  " ".join(allParams), "makespan time date", file=fd)
            for r in recordsWithParamDicts:
                print(r.insType,
                      " ".join(getattr(r, p) if getattr(r, p) else "NA" for p in instanceParameters),
                      "%d %d" % (r.m, r.k),
                      "%s %s" % (r.isBound, r.alg),
                      " ".join(r.params[k] if k in r.params else "NA" for k in allParams),
                      "%s %s %s" % (r.mkspan, r.time, r.date.strftime(dateformat)), file = fd)

        def displayFileContents(self, pred = lambda x: True, **kwargs):
            data = self.readRecords()
            self.printRecords(filter(pred, data), **kwargs)

        def getRecordsFromDatFile(self, file, questions):
            knownFields = [("instance", "insType"), ("m", "m"), ("k", "k"), ("isBound", "isBound"),
                           ("algorithm", "alg")] + [(p, p) for p in instanceParameters]
            resultFields = [ ("makespan", "mkspan"), ("time", "time"), ("date", "date") ]
            newQuestions = [(q._replace(params=dict(eq.split('=') for eq in q.params.split(":")) if q.params else {}), q) for q in questions]
            result = []
            with open(file, "r") as fd:
                reader = csv.DictReader(fd, delimiter=' ')
                for row in reader:
                    self.convertRow(row)
                    known = {p:row[p] for (p,z) in knownFields + resultFields}
                    for (p, z) in knownFields + resultFields:
                        del row[p]
                    row = {k:v for (k, v) in row.items() if v != "NA"}
                    for (q, origQ) in newQuestions:
                        if all(getattr(q, z) == known[p] for (p, z) in knownFields) and row == q.params:
                            result.append(self.Record(**origQ._asdict(), mkspan=known["makespan"], time=known["time"],
                                                      date = known["date"]))
                            break
            return result

    return Storage
    
            

        
